package diva.apps;
import diva.sketch.classification.FeatureSet;
import diva.sketch.features.FEUtilities;
import diva.sketch.recognition.TimedStroke;
import java.util.ArrayList;
import java.util.Iterator;

public class Generator {
    double[] _lengthMu;
    double[] _lengthSigma;
    ThetaVector[] _thetaMuVectors;
    double[] _thetaSigma;
    double[] _thetaMu;
    int _numPoints;
    double _threshold = 0.8; // probability has to be above the threshold

    FeatureSet[] _thetaList;//for sampling theta
    
    /**
     * Given a list of strokes each of which has 'numPoints' vertices,
     * compute mu and sigma for theta and segment length at each
     * interval.
     */
    public Generator (ArrayList strokes, int numPoints){
        _numPoints = numPoints;
        computeLength(strokes, numPoints);
        computeTheta(strokes, numPoints);
    }

    private void computeTheta(ArrayList strokes, int numPoints){
        int numStrokes = strokes.size();
        int numTs = numPoints-1;
        _thetaMuVectors = new ThetaVector[numTs];
        for(int i=0; i<strokes.size(); i++){
            TimedStroke s = (TimedStroke)strokes.get(i);
            if(i==0){//first stroke
                for(int j=0; j<s.getVertexCount()-1; j++){
                    _thetaMuVectors[j] = new ThetaVector();
                    _thetaMuVectors[j].x1 = s.getX(j);
                    _thetaMuVectors[j].y1 = s.getY(j);                    
                    _thetaMuVectors[j].x2 = s.getX(j+1);
                    _thetaMuVectors[j].y2 = s.getY(j+1);                    
                }
            }
            else {//concatenate vectors
                for(int j=0; j<s.getVertexCount()-1; j++){
                    double dx = _thetaMuVectors[j].x2-s.getX(j);
                    double dy = _thetaMuVectors[j].y2-s.getY(j);
                    _thetaMuVectors[j].x2 = s.getX(j+1) + dx;
                    _thetaMuVectors[j].y2 = s.getY(j+1) + dy;
                }
            }
        }
        _thetaMu = new double[numTs];        
        _thetaSigma = new double[numTs];
        for(int i=0; i< numTs; i++){//init
            _thetaMu[i] = 0;
            _thetaSigma[i] = 0;
        }
        double x, y, dx, dy, mag1, mag2, prod, theta;
        ThetaVector tv;
        for(int i=0; i<strokes.size(); i++){
            TimedStroke s = (TimedStroke)strokes.get(i);
            for(int j=0; j<s.getVertexCount()-1; j++){
                tv = _thetaMuVectors[j];
                dx = tv.x2-s.getX(j);
                dy = tv.y2-s.getY(j);
                x = s.getX(j+1)+dx;
                y = s.getY(j+1)+dy;
                prod = FEUtilities.dotProduct(tv.x1, tv.y1, tv.x2, tv.y2, x, y);
                mag1 = FEUtilities.distance(tv.x1, tv.y1, tv.x2, tv.y2);
                mag2 = FEUtilities.distance(s.getX(j), s.getY(j), s.getX(j+1), s.getY(j+1));
                theta = Math.acos(prod/(mag1*mag2));
                _thetaSigma[j] += Math.pow(theta, 2);
            }
        }
        for(int i=0; i< numTs; i++){//init
            _thetaSigma[i] /= numStrokes;
            _thetaMu[i] = 2*Math.PI-FEUtilities.computeAngle(_thetaMuVectors[i].x1, _thetaMuVectors[i].y1, _thetaMuVectors[i].x2, _thetaMuVectors[i].y2);
        }

        //**********************************************************//
        _thetaList = new FeatureSet[numStrokes];
        int k=0;
        for(Iterator iter = strokes.iterator(); iter.hasNext();){
            TimedStroke s = (TimedStroke)iter.next();
            //compute theta at each point along a stroke
            double xhead = s.getX(0);
            double yhead = s.getY(0);
            double xtail, ytail, angle, length;
            double[] thetas = new double[numTs];//we don't compute at the last point
            int j = 0;
            for(int i=1; i<numPoints; i++){
                xtail = s.getX(i);
                ytail = s.getY(i);
                angle = FEUtilities.computeAngle(xhead, yhead, xtail, ytail);
                if(Double.isNaN(angle)){// need to make sure it doesn't cause array out of bound
                    angle = thetas[i-1];
                    thetas[j] = angle;
                }
                else{
                    //since the origin of the coordinate system is at the
                    //upper left corner, we had to do angle conversion
                    angle = 2*Math.PI-angle;
                    thetas[j] = angle;
                }
                j++;
                xhead = xtail;
                yhead = ytail;
            }
            FeatureSet thetaFS = new FeatureSet(thetas);
            _thetaList[k] = thetaFS;
            k++;
        }
    }
    
    private void computeLength(ArrayList strokes, int numPoints){
        int numStrokes = strokes.size();
        FeatureSet[] lengthList = new FeatureSet[numStrokes];
        //if there are 9 points, t goes from 1 to 8.            
        int numTs = numPoints-1;
        int k=0;
        for(Iterator iter = strokes.iterator(); iter.hasNext();){
            TimedStroke s = (TimedStroke)iter.next();
            //compute length at each point along a stroke
            double xhead = s.getX(0);
            double yhead = s.getY(0);
            double xtail, ytail, angle, length;
            double[] lengths = new double[numTs];
            int j = 0;
            for(int i=1; i<numPoints; i++){
                xtail = s.getX(i);
                ytail = s.getY(i);
                length = FEUtilities.distance(xhead, yhead, xtail, ytail);
                lengths[j] = length;
                j++;
                xhead = xtail;
                yhead = ytail;
            }
            FeatureSet lengthFS = new FeatureSet(lengths);
            lengthList[k] = lengthFS;
            k++;
        }

        _lengthMu = new double[numTs];
        _lengthSigma = new double[numTs];        
        for(int i=0; i< numTs; i++){//init
            _lengthMu[i] = 0;            
            _lengthSigma[i] = 0;            
        }
        //compute mu for segment length
        for(int i=0; i< numStrokes; i++){
            FeatureSet lengthFs = lengthList[i];
            for(int j=0; j<numTs; j++){
                _lengthMu[j] += lengthFs.getFeature(j);
            }
        }
        for(int i=0; i<numTs; i++){
            _lengthMu[i] /= numStrokes;
        }
        //compute sigma
        for(int i=0; i< numStrokes; i++){
            FeatureSet lengthFs = lengthList[i];
            for(int j=0; j<numTs; j++){
                _lengthSigma[j] += Math.pow(lengthFs.getFeature(j)-_lengthMu[j], 2);
            }
        }
        for(int i=0; i<numTs; i++){
            _lengthSigma[i] /= numStrokes;
        }
    }
    
    class ThetaVector {
        public double x1;
        public double y1;
        public double x2;
        public double y2;        
    }

    private int getQuadrant(double theta){
        if((theta>=0) && (theta< Math.PI/2)){
            return 1;
        }
        else if((theta>= Math.PI/2) && (theta< Math.PI)){
            return 2;
        }
        else if((theta>= Math.PI) && (theta < (3/2)*Math.PI)){
            return 3;
        }
        else{
            return 4;
        }
    }
    
    /**
     * t1 and t2 are angles in radians (0-2pi)
     */
    private double genTheta(double t1, double t2){
        double theta = t1;
        if(getQuadrant(t1)==1 && getQuadrant(t2)==4){
            double delta = (2*Math.PI-t2)+t1;
            theta = t1-Math.random()*delta;
        }
        else if(getQuadrant(t1)==4 && getQuadrant(t2)==1){
            double delta = (2*Math.PI-t1)+t2;
            theta = t2-Math.random()*delta;
        }
        else{
            double delta = Math.abs(t1-t2);
            theta = Math.min(t1,t2)+Math.random()*delta;
        }
        return theta;
    }

    /*    
    private double genTheta(int interval){
        double mu = _thetaMu[interval];
        double sigma = _thetaSigma[interval];        
        //double p = (1-_threshold)*Math.random()+_threshold;
        //System.out.println("p="+p);
        //        double theta = gaussian(mu, sigma, p);
        double factor = (Math.random()>0.5)?1:-1;
        double theta = factor*0.1*Math.random()*sigma + mu;
        if(Double.isNaN(theta)){
            return mu;
        }
        return theta;
    }
    */
    
    public TimedStroke genStroke(){
        int size = _thetaList.length;
        int s1 = ((int)(Math.random()*1000))%size;
        int s2 = ((int)(Math.random()*1000))%size;
        FeatureSet fs1 = _thetaList[s1];
        FeatureSet fs2 = _thetaList[s2];

        TimedStroke stroke = new TimedStroke();
        stroke.addVertex(0f,0f,0);
        double prevx = 0f;
        double prevy = 0f;
        for(int i=0; i<_numPoints-1; i++){
            double theta = genTheta(fs1.getFeature(i), fs2.getFeature(i));
            //double theta = genTheta(i);
            double length = genLength(i);//_lengthMu[i];
            double x = length*Math.cos(theta);
            double y = length*Math.sin(theta);
            double relx = x+prevx;
            double rely = (-y)+prevy;
            //           System.out.println("\tnew("+(int)x+", "+(int)y+") prev("+ (int)prevx+", " + (int)prevy+") rel("+(int)relx+", "+(int)rely+")");
            stroke.addVertex((float)relx, (float)rely, 0);
            prevx = relx;
            prevy = rely;
        }
        //        System.out.println();
        return stroke;
    }

    private double genLength(int interval){
        double mu = _lengthMu[interval];
        double sigma = _lengthSigma[interval];
        double p = (1-_threshold)*Math.random()+_threshold;
        double length = gaussian(mu, sigma, p);
        /*
        double factor = (Math.random()>0.5)?1:-1;
        double length = factor*(1-_threshold)*Math.random()*sigma + mu;        
        */
        if(Double.isNaN(length)){
            return mu;
        }
        return length;
    }

    public double gaussian(double mu, double sigma, double probability){
        //AR: mu + or - the rest
        double val = mu+Math.sqrt(-2*sigma*sigma*Math.log(probability*Math.sqrt(2*Math.PI*sigma*sigma)));
        return val;
    }

}
