# datamodel.tcl: Read and parse a tree-structured data represetation.
# From Tycho, adapted for vanilla Tcl.
#
# $Id: datamodel.tcl,v 1.4 2000/05/02 00:01:28 johnr Exp $
#
# Author: John Reekie
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


####
# Procedures for parsing and printing out data models.
#


# Create the tim namespace
namespace eval ::tim {
    namespace export *
}

########################################################################
#### add
#
# Add a new item of type _type_. _name_ is the full name of the item.
# Following arguments are attribute-value pairs. If the attribute list
# is followed by a single list argument, that argument is taken to be a
# description of child items of this item, which will be parsed
# in the context of the newly-added item. Raise an error if the
# item already exists or if its parent context does not exist.
#
proc tim::add {ns type name args} {
    # Do lots of time-consuming error-checking
    if [::info exists ${ns}::_data_($name)] {
        # error "A $type \"$name\" already exists in $ns"
        puts "A $type \"$name\" already exists in $ns"
    }
    if ![::info exists ${ns}::_defaults_] {
	error "Unknown type \"$type\" in $ns"
    }
    set context [lreplace $name end end]
    if { $context != "" \
            && ![::info exists ${ns}::_data_($context)] } {
	error "Unknown context \"$context\" in $ns"
    }

    # Extract contents
    if { [llength $args] & 1 } {
    	set contents [lindex $args end]
    	set args [lreplace $args end end]
    } else {
    	set contents {}
    }

    # Store the data of this item
    set ${ns}::_data_($name) $args

    # Store the type of this item
    set ${ns}::_type_($name) $type

    # Add it to the list of children of its parent
    lappend ${ns}::_children_:[set type]($context) [lindex $name end]

    # Parse contents
    if { $contents != "" } {
        set _context_ $name
        [set ${ns}::_safeinterp_] eval [list namespace eval $ns $contents]
    }
}

########################################################################
#### attrdel
#
# Delete individual attributes of the item given by _name_.
# Each given attribute is deleted from the attribute list.
# If an attribute does not exist, then ignore it.
# Raise an error if the item does not exist.
#
proc tim::attrdel {ns name args} {
    _verify $ns $name

    # Get attributes
    array set attrs [set ${ns}::_data_($name)]
    
    # Delete each
    foreach a $args {
	if [::info exists attrs($a)] {
	    unset attrs($a)
	}
    }
    set ${ns}::_data_($name) [array get attrs]
}

########################################################################
#### attrget
#
# Get individual attributes of the item given by _name_.
# If only _name_ is given, return the attribute list of the item,
# _including_ type defaults. If _attr_ is given, return the 
# value of that attribute, returning the default if this attribute
# has not been set and there is a type default. Raise an error if the
# item does not exist, or if _attr_ does not exist.
#
proc tim::attrget {ns name {attr {}}} {
    _verify $ns $name

    # Get defaults
    set type [set ${ns}::_type_($name)]
    array set attrs [set ${ns}::_defaults_($type)]
    array set attrs [set ${ns}::_data_($name)]

    if { $attr == "" } {
        # Return all attributes
        array get attrs
    } else {
	# Return a single attribute
	if { ! [::info exists attrs($attr)] } {
	    error "Unknown attribute $attr of $type \"$name\"" in $ns"
	} ;#"
	return $attrs($attr)
    }
}

########################################################################
#### attrset
#
# Set individual attributes of the item given by _name_.
# If only _name_ is given, return the attribute list of the item,
# _not including_ type defaults. If additional arguments are given,
# change each attribute to the corresponding value. Raise an error
# if the item does not exist.
#
proc tim::attrset {ns name args} {
    _verify $ns $name

    # If no arguments, return data without defaults
    if { $args == "" } {
        return [set ${ns}::_data_($name)]
    }

    # Get current data
    array set attrs [set ${ns}::_data_($name)]
    
    # Merge the new data and return result
    array set attrs $args
    set ${ns}::_data_($name) [array get attrs]
}

########################################################################
#### clear
#
# Clear all data from the model.
#
proc tim::clear {ns} {
    # Unset the data of all items and the children of each type
    catch {unset ${ns}::_data_}
    catch {unset ${ns}::type}
    foreach type [types $ns] {
        catch {unset ${ns}::_children_:$type}
    }
}

########################################################################
#### cget
#
# Get a configuration option of a model.
#
proc ::tim::cget {ns option} {
    set ${ns}::_options_($option)
}

########################################################################
#### configure
#
# Configure a model. The configuration options are stored in
# the array _options, and default options are stored in
# _defaultoptions.
#
proc ::tim::configure {ns args} {
    # FIXME
    if { $args == "" } {
	# FIXME
	array get ${ns}::_options_
    } else {
	array set ${ns}::_options_ $args
	return [array get ${ns}::_options_]
    }
}

########################################################################
#### create
#
# Create a new model. The first argument is the name of the model,
# which is both the namespace the data is contained in and the access
# procedure for "object-oriented" access. This name is also returned
# as the result of this procedure. The only "type" defined is the
# configure{} psuedo-type.
#
proc ::tim::create {ns args} {
    set origns $ns

    # Create a new namespace and get the fully-qualified name
    set ns [uplevel "namespace eval $ns {namespace current}"]

    # Import all commands
    namespace eval $ns {namespace import -force ::tim::*} 

    # Create a safe interpreter for parsing
    set ${ns}::_safeinterp_ [uplevel #0 interp create]

    # Add the namespace to the safe interpreter
    [set ${ns}::_safeinterp_] eval "namespace eval $ns {}"

    # Create a procedure for "object-oriented" access to the
    # data.
    uplevel [subst {
	proc $ns {method args} {
	    namespace eval $ns \$method $ns \$args
	}
    }
    ]

    # Initialise the context for recursive parsing
    set ${ns}::_context_ {}

    # Create and link the configure{} proc into the safe interpreter
    namespace eval $ns [subst {
	proc __parse_configure {args} {
	    eval ::tim::configure $ns \$args
	}
    }
    ]
    [set ${ns}::_safeinterp_] alias ${ns}::configure ${ns}::__parse_configure

    # Evaluate options
    eval ${ns}::__parse_configure $args

    return $origns
}

########################################################################
#### define
#
# Define a new type. A model must have one or more types
# defined before it can be useful. Each
# type has a set of type defaults. Each
# type default gives the name and default value of an attribute
# that will be given to instances of that type. By convention,
# attributes have names beginning with a leading dash, but
# this is not required. Type defaults apply to individual models,
# and so can be changed without affecting other models of the same class.
# If the type has already been defined for this
# model, then the type default will be overridden with the new
# values.
#
proc ::tim::define {ns type args} {
    # Merge new defaults with existing default (if any)
    if [info exists ${ns}::_defaults_($type)] {
	array set temp ${ns}::_defaults_($type)
    }
    array set temp $args
    set ${ns}::_defaults_($type) [array get temp]

    # Create and define the parsing proc
    namespace eval $ns [_mkparseproc $ns $type]

    # Link the proc into the safe interpreter
    [set ${ns}::_safeinterp_] alias ${ns}::$type ${ns}::__parse_$type
}

########################################################################
#### delete
#
# Delete the item given by _name_. Raise an error if
# the item does not exist.  The item and all child items will
# be deleted.
#
proc tim::delete {ns name} {
    _verify $name
    set type [set ${ns}::_type_($name)]
    set context [lreplace $name end end]

    # Delete children
    foreach t [types $ns] {
	foreach child [match $ns $t $name] {
	    _delete $ns $name [concat $name $child] $t
	}
    }

    # Remove from the parent list
    set ${ns}::_children_:[set type]($context) \
            [::tycho::ldelete \
            [set ${ns}::_children_:[set type]($context)] \
            [lindex $name end]]

    # Delete data
    unset ${ns}::_data_($name)
}

# Helper method to delete children
proc tim::_delete {context name type} {
    # Delete children
    foreach t [types $ns] {
        foreach child [match $ns $t $name] {
            _delete $name [concat $name $child] $t
        }
    }
    # Delete the parent's child list
    set ${ns}::_children_:[set type]($context) {}

    # Delete data
    unset ${ns}::_data_($name)
}

########################################################################
#### describe
#
# Generate a TIM description of an item or of the model. If _name_ is
# not supplied, generate a description of the model. The description
# starts with an emacs-style header string that contains the mode
# and other relevant information. Each public variable listed in the
# protected variable <code>printable</code> is then printed as a
# *configure* command if it differs from its default value.
# Then the items in the model are printed, sorted by type.
# If the _name_ argument is supplied, generate a description
# of just that item. If the _prefix_ argument is also supplied,
# prepend the prefix to each line produced.
#
proc ::tim::describe {ns {name {}} {prefix {}}} {
    if { $name!= "" } {
	# Describe a single item
        _verify $ns $name
        _describe $ns $name $prefix
	
    } else {
	# Print header string
	# FIXME
	# set string [header]
	
	# Print public variables
	#foreach {opt dflt val} [eval concat [configure $ns]] {
	#    set v($opt) $val
	#    set d($opt) $dflt
	#}
	#foreach opt $printable {
	#    if { $v($opt) != $d($opt) } {
	#	append string "\n$prefix"
	#	append string [list configure $opt $v($opt)]
	#    }
	#}
	# FIXME
	foreach {opt val} [configure $ns] {
	    append string "\n$prefix"
	    append string [list configure $opt $val]
	}

	# Print items, sorted by type
	append string "\n"
	foreach type [tim::types $ns] {
	    # Describe each root item
	    foreach n [tim::match $ns $type] {
		 append string [tim::_describe $ns $n $prefix]
	    }
	}
	return $string
    }
}

########################################################################
#### destroy
#
# Destroya model.
#
proc ::tim::destroy {ns} {
    namespace delete $ns
    rename $ns {}
}

########################################################################
#### exists
#
# Test if the item given by _name_ is present in the model, returning
# 1 if it does and 0 if not. If the item does exists and _attribute_
# is given, test if the item has that attribute (or a default
# with that name exists), and return 1 if it does and 0 if it dies not.
#
proc tim::exists {ns name {attribute {}}} {
    if { $attribute == "" } {
	::info exists ${ns}::_data_($name)
    } else {
	if [::info exists ${ns}::_data_($name)] {
	    set type [set ${ns}::_type_($name)]
	    array set data [set ${ns}::_defaults_($type)]
	    array set data [set ${ns}::_data_($name)]
	    ::info exists data($attribute)
	} else {
	    return 0
	}
    }
}

########################################################################
#### get
#
# Return the names and data of items with names that match _pattern_,
# using glob-style matching. If _pattern_ is not supplied, get the
# names and data of all items. Note that a pattern such as
# <code>{foo *}</code> will return all descendents, not just
# children, of <code>foo</code>. The returned list is not
# be ordered.
#
proc tim::get {ns {pattern *}} {
    array get ${ns}::_data_ $pattern
}

########################################################################
#### match
#
# Get child items of the given _type_, optionally matching against
# attribute values. The returned list contains the local names
# of matched items, in the order in which the items were added
# to the model (or subsequently re-ordered with <b>reorder</b>).
# If the _context_ is given, get items within that context;
# otherwise, get top-level items. If _context_ is supplied and
# does not exist, raise an error. Any additional arguments are
# taken to be a list of attribute name-value pairs. If supplied,
# only items that have the given attributes with the given values
# will be returned.
#
proc tim::match {ns type {context {}} args} {
    # Adjust for presence of context argument
    if { [llength $args] & 1 } {
        set args [linsert $args 0 $context]
        set context {}
    } elseif { $context != "" && ![exists $ns $context] } {
        error "Unknown context \"$context\" in $ns"
    }
    # Get all children in order
    if [::info exists ${ns}::_children_:[set type]($context)] {
        set children [set ${ns}::_children_:[set type]($context)]
        if { $args == "" } {
            return $children
        }
    } else {
        return {}
    }
    # Filter by attribute
    set result {}
    foreach n $children {
        set name [concat $context $n]

        array set attrs [set ${ns}::_defaults_($type)]
        array set attrs [set ${ns}::_data_($name)]

    	set matching 1
    	foreach {attr val} $args {
	    if { ![::info exists attrs($attr)] || $attrs($attr) != $val } {
                set matching 0
                break
            }
        }
        if { $matching } {
            lappend result $n
        }
        unset attrs
    }
    return $result
}

########################################################################
#### names
#
# Return item names that match _pattern_. If _pattern_ is not supplied
# get the names of all items in the model. Note that this method matches
# item names against a pattern, so that a pattern such as
# <code>{foo *}</code> will return all descendants, not just children
# of <code>foo</code>. The returned list is not be ordered.
#
proc tim::names {ns {pattern *}} {
    array names ${ns}::_data_ $pattern
}

########################################################################
#### parse
#
# Read data from a string containing a TIM representation of the
# model. Each non-comment line must either start with the keyword
# <b>configure</b>. or with the name of a type defined for this
# model. The item name follows, then the attribute-value list,
# and then nested items within curly braces. If any item
# already exists, then an error is raised. If _context_ is supplied,
# then _data_ will be parsed relative to that item.
#
proc ::tim::parse {ns data {context {}}} {
    if { $context != "" && ![exists $context] } {
	error "Unknown context \"$context\" in $ns"
    }

    set saved [set ${ns}::_context_]
    set ${ns}::_context_ $context
    [set ${ns}::_safeinterp_] eval [list namespace eval $ns $data]
    set ${ns}::_context_ $saved
}

########################################################################
#### reorder
#
# Change the ordering of items of a given type. _type_ is
# the type of items to reorder. If _context_ is not present,
# then reorder items at the top level, otherwise reorder
# items within that context. _list_ is a list that
# must contain the same names returned by a call to
# <code>match type ?context?</code> but in any order.
# For example, items can be sorted alphabetically
# by a call of the form
# <pre>
#     $model reorder type [lsort [$model match type]]
# </pre>
# Future calls to <b>match</b> and <b>describe</b> will return
# items in the new order given by _list_.
#
proc tim::reorder {ns type {context {}} {list {}}} {
    # Adjust for presence of context argument
    if { $list == {} } {
        set list $context
        set context {}
    }
    # Check that the list contains the right names
    if { [lsort $list] != [lsort [match $type $context]] } {
        error "Child names do not match"
    }
    set ${ns}::_children_:[set type]($context) $list
}

########################################################################
#### type
#
# Return the type of the given item.
#
proc ::tim::type {ns name} {
    _verify $ns $name
    set ${ns}::_type_($name)
}

########################################################################
#### types
#
# Return a list of types defined for this model.
#
proc ::tim::types {ns} {
    array names ${ns}::_defaults_
}

#####################################################################
#####################################################################
####                     helper procedures                       ####

########################################################################
#### _describe
#
# Return a string describing the item given by _name_
# in the TIM format. _prefix_, if supplied, is prepended to each line.
#
proc ::tim::_describe {ns name {prefix {}}} {
    set type [set ${ns}::_type_($name)]

    # Start string with type followed by local name
    set string "$prefix$type [lindex $name end]"

    # Print attributes
    # This is fast:
    eval lappend string [set ${ns}::_data_($name)]

    # This looks better:
    # It's also broken! breakLines adds a back-slash that makes
    # Tcl back-slash escape the whole string. Tcl is such a piece
    # of shit.
    #append string " " [string trimleft [::tycho::breakLines \
	    #        [@scope $scope set data($name)] \
	    #        -length 60 \
	    #        -substring " -" \
	    #        -prefix $prefix] " "]

    # Print children if I have any
    if { [info exists ${ns}::_children_($name)] } {
	set cd [_describechildren $name "$prefix    "]
	append string " \{\n$cd$prefix\}"
    }
    return $string\n
}

########################################################################
#### _describechildren
#
# Return a string describing the children of the item given by _name_
# in the TIM format. The _prefix_, if supplied, is prepended to each line.
# Enclosing braces are not printed.
#
proc ::tim::_describechildren {ns name {prefix {}}} {
    set string ""
    foreach t [tim::types $ns] {
	foreach child [tim::match $ns $t $name] {
	    append string [_describe $ns [concat $name $child] $prefix]
	}
    }
    return $string
}

########################################################################
#### _mkparseproc
#
# Create a proc to parse an item
#
proc ::tim::_mkparseproc {ns type} {
    # Proc header
    set proc "proc ${ns}::__parse_$type {tail args} \{\n"

    # Proc body
    append proc "    set type $type\n"

    # Declare various data global
    append proc {
	variable _context_
    }
    # Cache the namespace
    append proc {
	set ns [namespace current]
    }
    # Generate the full name of the item
    append proc {
	set name [concat $_context_ $tail]
    }
    # Check whether the item already exists
    append proc {
	if [::info exists ${ns}::_data_($name)] {
	    # error "A $type \"$name\" already exists in $ns"
	    puts "A $type \"$name\" already exists in $ns"
	}
    }
    # Extract contents
    append proc {
	if { [llength $args] & 1 } {
	    set contents [lindex $args end]
	    set args [lreplace $args end end]
	} else {
	    set contents {}
	}
    }
    # Set the data of this item
    append proc {
	set ${ns}::_data_($name) $args
    }
    # Remember its type
    append proc {
	set ${ns}::_type_($name) $type
    }
    # Add the item tail to the list of children of its parent
    append proc {
	lappend ${ns}::_children_:[set type]($_context_) $tail
    }
    # Parse contents. Use the safe interpreter to ensure safety.
    append proc {if { $contents != "" }}
    set t {
	set saved [set ${ns}::_context_]
	set ${ns}::_context_ $name
    }
    append t "[set ${ns}::_safeinterp_] \
	    eval namespace \"eval $ns \{\$contents\n\}\""
    append t {
	set ${ns}::_context_ $saved
    }
    append proc " \{$t\}"

    # Complete the proc and return it
    append proc "\}"
}

########################################################################
#### _verify
#
# Verify that the item given by _name_ exists, raising an
# error if it does not.
#
proc tim::_verify {ns name} {
    if ![::info exists ${ns}::_data_($name)] {
    	error "Unknown item \"$name\" in $ns"
    }
}






