# divaTree.tcl: Utilities related to the Diva tree
#
# $Id: divaTree.tcl,v 1.25 2000/05/02 00:01:28 johnr Exp $
#
# Copyright (c) 1998-2000 The Regents of the University of California.
# All rights reserved. See the file COPYRIGHT for details.
#


### Assumed global variables:
# debug: print debug information if on
# silent: don't print anything if on


# Patterns for all binary files types
set binaryFileTypes {
    *.class
    *.doc
    *.fm
    *.gif
    *.gz
    *.jar
    *.jpg
    *.ppt
    *.tar
    *.vsd
    *.zip
    *.xar
    *.xpr
}


#######################################################################
#### addCopyrights
#
# Copy the file COPYRIGHT from the Diva root
# into every directory of the Diva tree. This is used to make
# sure the copyright notice is plastered all over the distribution.
#
proc addCopyrights {DIVA dir} {
    global debug silent

    if { $DIVA != $dir } {
	file copy -force [file join $DIVA COPYRIGHT] [file join $dir]
# MMS
#	file copy -force [file join $DIVA copyright.html] [file join $dir]
#
    }
    foreach d [glob -nocomplain [file join $dir *]] {
	if [file isdirectory $d] {
	    addCopyrights $DIVA [file join $dir $d]
	}
    }
}

#######################################################################
#### divaConfigure
#
# Given a diva root, attempt to "configure" the makefile system.
# This is our poor person's configure, but it's good enough. The
# first argument is any diva root, the second is the one being
# configured. Return 1 if successful, 0 if not. Be verbose about it.
# The relevant data is in the arrays at the top of this file.
#
proc divaConfigure {DIVA {path {}}} {
    global silent debug env

    if { $path == "" } {
	set path $DIVA
    }

    #
    # CD into the path and run ./configure
    #
    cd $path
    puts -nonewline "Configuring for your system... "
    if [ catch {exec ./configure > tmpfile} result ] {
	puts "configure failed"
	return 0
    }

    puts "configure succeded"
    return 1
}

#######################################################################
#### divaPackages
#
# Given a list of package names or makefile variables, return
# the list of existing packages given by those names and variables. 
#
proc divaPackages {DIVA args} {
    global debug silent
    if {$args == "" } {
	set args "PACKAGES AUXILIARY_PACKAGES"
    }

    set packages {}
    while { $args != "" } {
	set p [lindex $args 0]
	set args [lreplace $args 0 0]
	if { [string first . $p] < 0 } {
	    # No period, so it's in a makefile
	    if $debug {
		puts "Reading makefile variable \"$p\""
	    }
	    if [catch {getMakevar $p} new] {
		puts "ERROR: make variable \"$p\" not found"
		continue
	    }
	    set args [concat $args $new]
	} else {
	    # Check the directory
	    set dir [eval file join $DIVA [lreplace [split $p .] 0 0]]
	    if ![file exists $dir] {
		puts "Package directory \"$dir\" does not exist"

	    } else {
		lappend packages $p
	    }
	}
    }
    return $packages
}

#######################################################################
#### divaPath
# Given the root of the Diva tree and an absolute path, return
# the relative path from the diva root to the given path. For this
# to work, both paths must be in the "canonical" form returned
# by [pwd].
# 
proc divaPath {root dir} {
    set root [file split $root]
    set dir [file split $dir]
    while { $root != "" && $dir != "" \
	    && [lindex $root 0] == [lindex $dir 0] } {
	set root [lreplace $root 0 0]
	set dir [lreplace $dir 0 0]
    }
    return [eval file join diva $dir]
}   

#######################################################################
#### relativePath
# Given two absolute paths, return the path of the second relative
# to the first. For this to work, both paths must be in the
# "canonical" form returned by [pwd].
# 
proc relativePath {base path} {
    set base [file split $base]
    set path [file split $path]
    while { $base != "" && $path != "" \
	    && [lindex $base 0] == [lindex $path 0] } {
	set base [lreplace $base 0 0]
	set path [lreplace $path 0 0]
    }
    set path [eval file join $path]
    while { $base != "" } {
	set path [file join .. $path]
	set base [lreplace $base 0 0]
    }
    return $path
    
}   

#######################################################################
#### divaRoot
# Return the root of the Diva tree, relative to the given
# directory. This proc is somewhat random, but will work
# unless you try hard to fool it. If there's a makefile in
# this directory, it reads the makefile and extracts
# the DIVA variable from it. Otherwise, it assumes that this
# directory is in the documentation tree of Diva and
# goes up to the parent of the lowest directory name "doc."
# Then it check for the presence of a file named README.txt
# and the following directories: bin, doc, mk, canvas, graph.
# 
proc divaRoot {dir} {
    global debug silent
    
    if ($debug) {
	puts "divaRoot: getting root of directory $dir"
    }
    set cwd [pwd]
    cd $dir

    set root ""
    # Read the makefile if there is one
    if [file exists makefile] {
	parseMakefile makefile

	# Note that this will work even if DIVA is in the
	# environment. So unset it first
	catch {unset env(DIVA)}
	if [catch {set root [getMakevar DIVA]} result] {
	    error "Directory \"dir\" has a makefile with no DIVA"
	}
	set root [file join [pwd] $root]
	cd $root
	set root [pwd]

    } else {
	# Look for a doc directory
	set path [file split $dir]
	while {[lindex $path end] != "doc" && $path != ""} {
	    set path [lreplace $path end end]
	}
	if { $path == "" } {
	    error "Unable to find Diva directory \"doc\" above \"$dir\""
	}
	set root [eval file join [lreplace $path end end]]
    }
    # Now check if this looks like the right thing
    #set files "README.txt bin doc mk canvas graph"
    #cd $root
    #if { [eval glob $files] != $files } {
    #	error "The directory \"$root\" doesn't look like Diva"
    #}
    #cd $cwd
    if ($debug) {
	puts "EXIT divaRoot: root of $dir is $root"
    }
    cd $cwd
    return $root
}

#######################################################################
#### checkoutDiva
#
# Check out diva into a given directory. The first argument is the current diva
# root, and the second is the absolute or relative path to check-out into.
#
proc checkoutDiva {DIVA path} {
    global silent debug

    cd $DIVA

    # Figure out where the repository is
    set fd [open [file join CVS Root]]
    set repository [string trim [read $fd] " \n"]
    close $fd

    # Create the build directory
    if ![file exists $path] {
	file mkdir $path
    }
    cd $path
    if !$silent {
	puts "Deleting \"[file join [pwd] diva]\""
    }
    file delete -force diva

    # Check it out
    if !$silent {
	puts "Checking out Diva into \"[pwd]\""
	puts "exec cvs -d $repository co diva"
	puts "(You may need to type your CVS password now.)"
    }
    # BY THE WAY, if you need to type a password just type
    # it after the above line and it will work...
    catch {exec cvs -d $repository co diva} result

    return $result
}

#######################################################################
#### exportDiva
#
# Export diva into a given directory. The first argument is the current diva
# root, and the second is the absolute or relative path to export into.
#
proc exportDiva {DIVA path} {
    global silent debug

    cd $DIVA

    # Figure out where the repository is
    set fd [open [file join CVS Root]]
    set repository [string trim [read $fd] " \n"]
    close $fd

    # Create the build directory
    if ![file exists $path] {
	file mkdir $path
    }
    cd $path
    if !$silent {
	puts "Deleting \"[file join [pwd] diva]\""
    }
    file delete -force diva

    # Check it out
    if !$silent {
	puts "Exporting Diva into \"[pwd]\""
	puts "exec cvs -d $repository export -D today diva"
	puts "(You may need to type your CVS password now.)"
    }
    # BY THE WAY, if you need to type a password just type
    # it after the above line and it will work...
    catch {exec cvs -d $repository export -D today diva} result

    return $result
}

#######################################################################
#### updateDiva
#
# Update diva into a given directory. The first argument is the current diva
# root, and the second is the diva root to update in.
#
proc updateDiva {DIVA path} {
    global silent debug

    # Check it out
    cd $path
    if !$silent {
	puts "Updating Diva into \"[pwd]\""
	puts "exec cvs -q update -d -P"
	puts "(You may need to type your CVS password now.)"
    }
    # BY THE WAY, if you need to type a password just type
    # it after the above line and it will work...
    catch {exec cvs -q update -d -P} result

    return $result
}

#######################################################################
#### isBinaryFile
# Test if the given file is binary.
# 
proc isBinaryFile {file} {
    global binaryFileTypes
    set file [file tail $file]
    foreach p $binaryFileTypes {
	if [string match $p $file] {
	    return 1
	}
    }
    return 0
}

#######################################################################
#### pkgDirs
# Gives the diva root and arguments that are package names,
# return the list of the absolute paths to those packages.
# 
proc pkgDirs {DIVA args} {
    global debug silent

    set result ""
    foreach p $args {
	set dir [eval file join $DIVA [lreplace [split $p .] 0 0]]
	if { [file exists $dir] } {
	    lappend result $dir
	}
    }
    return $result
}

#######################################################################
#### windowWasher
#
# Make a copy of the tree, convert it to Windows format,
# and zip it.
# 
proc windowWasher {DIVA zipfile} {
    global debug silent

    if [file exists _ziptemp_] {
	file delete -force _ziptemp_
    }
    file mkdir _ziptemp_
    file copy $DIVA [file join _ziptemp_ diva]
    cd _ziptemp_
    if !$silent {
	puts "Munging files in [pwd]"
    }
    winMunge
    if !$silent {
	puts "zip -q -r -n .jpg:.gif $zipfile diva"
    }
    exec zip -q -r -n .jpg:.gif $zipfile diva
    exec mv $zipfile ..
    cd ..
    file delete -force _ziptemp_
}
proc winMunge {} {
    foreach f [glob -nocomplain *] {
	if [file isdirectory $f] {
	    cd $f
	    winMunge
	    cd ..
	} elseif ![isBinaryFile $f] {
	    ### The slow way: catch {exec unix2dos $f > tmpfile}
	    ###               file rename -force tmpfile $f

	    ### The fast way: (?)
	    set fd [open $f]
	    set text [read $fd]
	    close $fd
	    set fd [open $f w]
	    fconfigure $fd -translation crlf
	    puts $fd $text
	    close $fd
	}
    }
}






